/**
 * @file
 * Make book nav trees.
 *
 * Works best with the Bootstrap theme.
 */

(function ($, Drupal) {
  "use strict";
  Drupal.skillingSettings = Drupal.skillingSettings || {};
  // Flag to track initialization.
  Drupal.skillingSettings.shownBookTrees = false;
  Drupal.behaviors.showBookTrees = {
    attach: function (context, settings) {
      if (Drupal.skillingSettings.shownBookTrees) {
        return;
      }
      Drupal.skillingSettings.shownBookTrees = true;
      $(document).ready(function() {
        let throbber = $("#block-skillingbooknavigation .skilling-throbber-container");
        throbber.show();
        // Load and show data.
        // Trigger Ajax download and display.
        $.ajax({
          url: Drupal.url("skilling/load-book-navs"),
          type: 'POST',
          dataType: 'json',
          // data: {
          //   "csrfToken": drupalSettings.csrfToken,
          //   "sessionId": drupalSettings.sessionId,
          // },
          success: function (result) {
            if (result.status !== undefined) {
              if (result.status.toLowerCase() === "error") {
                swal(result.errorMessage);
                return;
              }
            }
            for(var treeDataSet of result) {
              Drupal.behaviors.showBookTrees.showBookTree(treeDataSet);
            }
            $(throbber).hide();
          },
          fail: (function (jqXHR, textStatus) {
            $(throbber).hide();
            let message = Drupal.t(
                "Skilling book tree load fail @text. Please send a screen shot of your "
                + "entire browser window, including URL, to someone.",
                {'@text': textStatus}
            );
            swal(message);
          })
        });
        Drupal.skillingSettings.bookNavsDownloaded = true;
      });
    },
    showBookTree: function (treeDataSet) {
      //Show the tree in a container that includes the book id as part of its id.
      //Ids are set up in SkillingBookNavBlock->build().
      let containerId = '#skilling-book-nav-' + treeDataSet.bid;
      let $container = $(containerId);
      //Flag to avoid response to activating node when initializing.
      let initializing = true;
      let tree = $container.fancytree({
        source: treeDataSet.tree.data,
        activate: function(event, data) {
          if ( ! initializing ) {
            let nid = data.node.key;
            if (nid) {
              // Is there an alias available?
              let alias = data.node.data.alias;
              let path = alias ? alias : "node/" + nid;
              let fullPath = Drupal.behaviors.skillingUtilities.getRootUrl() + path;
              window.location.href = fullPath;
            }
          }
        },
        autoActivate: false,
        icon: false
      });
      //Select tree item for the current Drupal node.
      let path = drupalSettings.path.currentPath;
      if ( path.indexOf("node/") === 0 ) {
        let nid = path.slice(5);
        if ( ! isNaN(nid) ) {
          //Look up the node in the tree.
          let node = tree.fancytree("getTree").getNodeByKey(nid);
          //Found it? Might not, is there are multiple trees on the page.
          if ( node ) {
            node.setSelected(true);
            // Expand all parents, but don't scroll node into view,
            // else messes up expected browser back button behavior.
            node.makeVisible({scrollIntoView: false});
          }
        }
      }
      initializing = false;
      //Erase the loading throbber.
      $container.find(".skilling-throbber-container").remove();

    }
  }
})(jQuery, Drupal);
