<?php

namespace Drupal\skilling\Controller;

use Drupal\Core\Controller\ControllerBase;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Drupal\Core\Messenger\MessengerInterface;
use Drupal\skilling\Timeline;
use Drupal\node\NodeInterface;

/**
 * Class CalendarPlannerController.
 */
class CalendarPlannerController extends ControllerBase {

  /**
   * Drupal\Core\Messenger\MessengerInterface definition.
   *
   * @var \Drupal\Core\Messenger\MessengerInterface
   */
  protected $messenger;

  /**
   * Timeline service.
   *
   * @var \Drupal\skilling\Timeline
   */
  protected $timelineService;

  /**
   * Constructs a new CalendarPlannerController object.
   *
   * @param \Drupal\Core\Messenger\MessengerInterface $messenger
   * @param \Drupal\skilling\Timeline $timelineService
   */
  public function __construct(
    MessengerInterface $messenger,
    Timeline $timelineService
  ) {
    $this->messenger = $messenger;
    $this->timelineService = $timelineService;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('messenger'),
      $container->get('skilling.timeline')
    );
  }

  /**
   * Make calendar planner.
   *
   * @param \Drupal\node\NodeInterface $node
   *   The calendar.
   * @param \Drupal\skilling\Controller\string $startDate
   *   When the planner should start.
   *
   * @return array Return Hello string.
   *   Return Hello string.
   */
  public function makePlanner(NodeInterface $node, string $startDate) {
    $timestamp = strtotime($startDate);
    $planningCalendar = $this->timelineService->makePlanningTimelineForCalendar($node, $timestamp);
    $renderable = [
      '#theme' => 'planning_calendar',
      '#calendar' => $planningCalendar,
      '#cache' => [
        'max-age' => 0,
      ],
    ];
    return $renderable;
  }

}
