<?php

namespace Drupal\skilling\Plugin\HistoryItem;

use Drupal\skilling\Plugin\HistoryItemBase;

/**
 * Provides a suggestion history event.
 *
 * @SkillingHistoryItem(
 *  id = "skilling_create_suggestion_history_event",
 *  event_type_name = "create_suggestion",
 *  label = @Translation("Suggestion created"),
 *  description = @Translation("Student submits a suggestion."),
 * )
 */
class CreateSuggestion extends HistoryItemBase {

  /**
   * {@inheritdoc}
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\skilling\Exception\SkillingException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  public function getDetailsRenderable() {
    $suggestion = $this->extractDeets();
    $result = [
      '#theme' => 'history_created_suggestion_deets',
      '#suggestion' => $suggestion,
    ];
    return $result;
  }

  /**
   * {@inheritdoc}
   */
  public function getDetailsXml() {
    $suggestion = $this->prepareForXml($this->extractDeets());
    $result = "<suggestion-created>$suggestion</submission-created>";
    return $result;
  }

  /**
   * Extract the details into separate variables.
   *
   * @return array
   *   Deets in variables.
   */
  protected function extractDeets() {
    $deets = $this->historyEventNode->get('field_details')->value;
    // Unserialize it.
    $deets = unserialize($deets);
    $suggestion = $deets['suggestion'];
    return $suggestion;
  }

}
