<?php

namespace Drupal\skilling\Plugin\HistoryItem;

use Drupal\Component\Utility\Html;
use Drupal\skilling\Exception\SkillingNotFoundException;
use Drupal\skilling\Plugin\HistoryItemBase;

/**
 * Provides a login history event.
 *
 * @SkillingHistoryItem(
 *  id = "skilling_login_history_event",
 *  event_type_name = "login",
 *  label = @Translation("Login"),
 *  description = @Translation("User login."),
 * )
 */
class Login extends HistoryItemBase {
  /*
   * No event details.
   */

  /**
   * Is the event to be logged in history?
   *
   * Only log for students.
   *
   * @return bool
   *   Whether to log the event.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   * @throws \Drupal\skilling\Exception\SkillingNotFoundException
   */
  public function isLogged() {
    /** @var \Drupal\Core\Entity\EntityTypeManagerInterface $entityTypeManager */
    $entityTypeManager = \Drupal::service('entity_type.manager');
    /** @var \Drupal\user\UserInterface $loggingInUser */
    $loggingInUser = $entityTypeManager->getStorage('user')
      ->load($this->loggingInUserId);
    if (!$loggingInUser) {
      // Should not happen.
      throw new SkillingNotFoundException(
        Html::escape('Uid unknown: ') . $uid,
        __FILE__,
        __LINE__
      );
    }
    $isStudent = $loggingInUser->hasRole('student');
    return $isStudent;
  }

}
