<?php

namespace Drupal\skilling\Plugin\HistoryItem;

use Drupal\skilling\Plugin\HistoryItemBase;
use Drupal\skilling\SkillingConstants;

/**
 * Provides a save reflect note history event.
 *
 * @SkillingHistoryItem(
 *  id = "skilling_save_reflect_note_history_event",
 *  event_type_name = "save_reflect_note",
 *  label = @Translation("Reflect note saved"),
 *  description = @Translation("Student saves a reflect note."),
 * )
 */
class SaveReflectNote extends HistoryItemBase {

  /**
   * {@inheritdoc}
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\skilling\Exception\SkillingException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  public function getDetailsRenderable() {
    list($nidNodeWithReflectTag, $identifier, $note) = $this->extractDeets();
    // Look up title of node the reflect note is in.
    $nodeReflectNoteIsInTitle = $this->t('Unknown');
    /** @var \Drupal\Core\Entity\EntityTypeManagerInterface $entityTypeManager */
    $entityTypeManager = \Drupal::service('entity_type.manager');
    /** @var \Drupal\node\NodeInterface $nodeReflectNoteIsIn */
    $nodeReflectNoteIsIn = $entityTypeManager->getStorage('node')
      ->load($nidNodeWithReflectTag);
    if ($nodeReflectNoteIsIn) {
      $nodeReflectNoteIsInTitle = (string) $nodeReflectNoteIsIn->getTitle();
    }
    $result = [
      '#theme' => 'history_save_reflect_note_deets',
      '#title_node_reflect_note_in' => $nodeReflectNoteIsInTitle,
      '#identifier' => $identifier,
      '#note' => $note,
    ];
    return $result;
  }

  /**
   * {@inheritdoc}
   */
  public function getDetailsXml() {
    list($nidNodeWithReflectTag, $identifier, $note)
      = $this->prepareForXml($this->extractDeets());
    $result = "<reflect-note-saved nid-with-reflect-tag=\"$nidNodeWithReflectTag\" "
      . "identifier=\"$identifier\">$note</reflect-note-saved>";
    return $result;
  }

  /**
   * Extract the details into separate variables.
   *
   * @return array
   *   Deets in variables.
   */
  protected function extractDeets() {
    $deets = $this->historyEventNode->get(SkillingConstants::FIELD_DETAILS)->value;
    // Unserialize it.
    $deets = unserialize($deets);
    $nidNodeWithReflectTag = $deets['nid-with-reflect-tag'];
    $identifier = $deets['identifier'];
    $note = $deets['note'];
    return [
      $nidNodeWithReflectTag,
      $identifier,
      $note,
    ];
  }

}
