<?php

namespace Drupal\skilling\Plugin\HistoryItem;

use Drupal\skilling\Plugin\HistoryItemBase;

/**
 * Provides a submission-graded history event.
 *
 * @SkillingHistoryItem(
 *  id = "skilling_submission_graded_history_event",
 *  event_type_name = "submission_graded",
 *  label = @Translation("Submission graded"),
 *  description = @Translation("Grader grades a submission."),
 * )
 */
class SubmissionGraded extends HistoryItemBase {

  /**
   * {@inheritdoc}
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\skilling\Exception\SkillingException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  public function getDetailsRenderable() {
    list($studentUid, $exerciseNid, $graderUid, $complete, $overall, $responses) = $this->extractDeets();
    $result = [
      '#theme' => 'history_submission_graded_deets',
      '#complete' => $complete,
      '#overall' => $overall,
      '#responses' => $responses,
    ];
    // Look up exercise title.
    /** @var \Drupal\Core\Entity\EntityTypeManagerInterface $entityTypeManager */
    $entityTypeManager = \Drupal::service('entity_type.manager');
    $exercise = $entityTypeManager->getStorage('node')
      ->load($exerciseNid);
    if ($exercise) {
      $result['#exerciseNid'] = $exerciseNid;
      $result['#exerciseTitle'] = $exercise->get('title');
    }
    /** @var \Drupal\skilling\SkillingCurrentUser $currentUser */
    $currentUser = \Drupal::service('skilling.skilling_current_user');
    if ($currentUser->isAdministrator()) {
      // Admins get more stuff in their deets.
      /** @var \Drupal\skilling\SkillingUserFactory $userFactory */
      $userFactory = \Drupal::service('skilling.skilling_user_factory');
      $student = $userFactory->makeSkillingUser($studentUid);
      $grader = $userFactory->makeSkillingUser($graderUid);
      $result['#studentUid'] = $studentUid;
      $result['#studentName'] = $student->getAccountName();
      $result['#graderUid'] = $graderUid;
      $result['#graderName'] = $grader->getAccountName();
    }
    return $result;
  }

  /**
   * {@inheritdoc}
   */
  public function getDetailsXml() {
    list($studentUid, $exerciseNid, $graderUid, $complete, $overall, $responses)
      = $this->prepareForXml($this->extractDeets());
    $result = '<submission-graded '
      . 'student="' . $studentUid . '"'
      . 'grader="' . $graderUid . '"'
      . 'exercise="' . $exerciseNid . '"'
      . 'complete="' . $complete . '"'
      . 'overall="' . $overall . '"'
      . '>';
    $result .= '<responses>';
    foreach ($responses as $rubricItemNid => $itemResponses) {
      $result .= '<rubric-item id="' . $rubricItemNid . '">';
      foreach ($itemResponses as $itemResponse) {
        $result .= '<response id="' . $itemResponse . '"/>';
      }
      $result .= '</rubric-item>';
    }
    $result .= '</responses>';
    $result .= '</submission-graded>';
    return $result;
  }

  /**
   * Extract the details into separate variables.
   *
   * @return array
   *   Deets in variables.
   */
  protected function extractDeets() {
    $deets = $this->historyEventNode->get('field_details')->value;
    // Unserialize it.
    $deets = unserialize($deets);
    $studentUid = $deets['student'];
    $exerciseNid = $deets['exercise'];
    $graderUid = $deets['grader'];
    $complete = $deets['complete'];
    $overall = $deets['overall'];
    $responses = $deets['responses'];
    return [
      $studentUid,
      $exerciseNid,
      $graderUid,
      $complete,
      $overall,
      $responses,
    ];
  }

  /**
   * Is the event to be logged in history?
   *
   * Always log.
   *
   * @return bool
   *   Whether to log the event.
   */
  public function isLogged() {
    return TRUE;
  }

}
