<?php

namespace Drupal\skilling\Plugin\views\filter;

use Drupal\views\Plugin\views\display\DisplayPluginBase;
use Drupal\views\Plugin\views\filter\InOperator;
use Drupal\views\ViewExecutable;

/**
 * Filters by history event type.
 *
 * @ingroup views_filter_handlers
 *
 * @ViewsFilter("history_event_type_filter")
 */
class FilterHistoryEventType extends InOperator {

  /**
   * {@inheritdoc}
   */
  public function init(ViewExecutable $view, DisplayPluginBase $display, array &$options = NULL) {
    parent::init($view, $display, $options);
    $this->definition['options callback'] = array($this, 'generateOptions');
  }

  /**
   * Override query so no filtering if the user doesn't select any options.
   */
  public function query() {
    if (!empty($this->value)) {
      parent::query();
    }
  }

  /**
   * Skip validation if no options have been chosen.
   */
  public function validate() {
    if (!empty($this->value)) {
      parent::validate();
    }
  }

  /**
   * Helper function that generates the options.
   *
   * @return array
   *   Event type options.
   */
  public function generateOptions() {
    $itemManager = \Drupal::service('plugin.manager.skilling.history_item');
    $historyPluginDefinitions = $itemManager->getDefinitions();
    $options = [];
    foreach ($historyPluginDefinitions as $pluginId => $historyPluginDefinition) {
      $eventType = $historyPluginDefinition['event_type_name'];
      $label = $historyPluginDefinition['label'];
      $options[$eventType] = $label;
    }
    return $options;
  }

}
